import React from 'react'           // React to manage the GUI
import XLSX from 'xlsx'             // xlsx to read and write Excel XLSX Workbooks
import Player from './player'       // Everything that has to do with players
import Match from './match'         // Everything that has to do with matches
import Excel from './excel'         // Helper files to create Excel files

const FILTER_OFF = 'Alle'

/** Main application */
class App extends React.Component {
  /** 
   * Constructor
   * Defines the state and binds all 'this' in all functions to the object.
   */
  constructor () {
    super()

    // Define the state
    this.state = {
      player: Player.state,
      match: Match.state,
      worksheets: {
        'PlayerList': null,
        'Calendar': null
      },
      filters: {
        date: null,
        place: null,
        category: null
      }
    }

    // Bind 'this' to functions
    this.handlePlayerList = this.handlePlayerList.bind(this)
    this.handleCalendar = this.handleCalendar.bind(this)
    this.generatePlayerList = this.generatePlayerList.bind(this)
    this.generateCalendar = this.generateCalendar.bind(this)
    this.filterData = this.filterData.bind(this)
    this.generateSchedule = this.generateSchedule.bind(this)
    this.generatePayTable = this.generatePayTable.bind(this)

    this.unpaid = 'off'
  }

  calculateMatchFilters () {
    const dates = {}
    const places = []
    const categories = []
    this.state.match.matches.forEach((item) => {
      if (!!item.DatumString & !dates.hasOwnProperty(item.DatumString)) {
        dates[item.DatumString] = item.Datum
      }
      if (!!item.Ort & !places.includes(item.Ort)) {
        places.push(item.Ort)
      }
      if (!!item.Konkurrenz & !categories.includes(item.Konkurrenz)) {
        categories.push(item.Konkurrenz)
      }
    })
    const match = { ...this.state.match, dates, places, categories }
    this.setState({ match })
  }

  handlePlayerList (event) {
    const file = this.playerList.files[0]
    this.readWorkbook(file, this.generatePlayerList)
  }

  handleCalendar (event) {
    const file = this.calendar.files[0]
    this.readWorkbook(file, this.generateCalendar)
  }

  readWorkbook (file, callback) {
    const reader = new FileReader()
    reader.onload = (e) => {
      const data = e.target.result
      const workbook = XLSX.read(data, {type: 'binary', cellDates: true})
      console.log('Workbook after read:', workbook)
      if (workbook.SheetNames.length !== 1) {
        throw Error(`Expected only one worksheet in the file, but found ${workbook.SheetNames.length}.`)
      }
      const worksheet = XLSX.utils.sheet_to_json(workbook.Sheets[workbook.SheetNames[0]], { header: 1, raw: true })
      callback(worksheet)
    }
    reader.readAsBinaryString(file)
  }

  generatePlayerList (worksheet) {
    console.log('About to read the player list.')
    const worksheets = { ...this.state.worksheets }
    worksheets['PlayerList'] = worksheet
    this.setState({ worksheets })

    if (worksheet[4].length !== 32 & worksheet[3][0] !== 'Konkurrenz' & worksheet[3][31] !== 'bezahlt') {
      throw Error('Wrong file structure.')
    }
    const players = worksheet.slice(4, worksheet.length).map((playerData) => new Player.Player(playerData))
    const player = { ...this.state.player }
    player.players = players
    this.setState({ player })
    this.filterData()
    console.log('State after generating player list:', this.state)
  }

  generateCalendar (worksheet) {
    console.log('About to read the calendar.')
    const worksheets = { ...this.state.worksheets }
    worksheets['Calendar'] = worksheet
    this.setState({ worksheets })

    if (worksheet[2].length !== 8) {
      throw Error('Wrong file structure.')
    }
    const matches = worksheet.slice(2, worksheet.length).map((matchData) => new Match.MatchClass(matchData))
    const match = { ...this.state.match }
    match.matches = matches
    this.setState({ match })
    this.calculateMatchFilters()
    this.filterData()
    console.log('State after generating calendar:', this.state)
  }

  filterData () {
    const filters = {
      date: this.date.value !== FILTER_OFF ? this.date.value : null,
      place: this.place.value !== FILTER_OFF ? this.place.value : null,
      category: this.category.value !== FILTER_OFF ? this.category.value : null,
    }
    console.log('New filter settings:', filters)
    this.setState({ filters })

    const match = { ...this.state.match }
    match.filtered = match.matches.filter((match) => {
      const matchDate = new Date(match.Datum)
      matchDate.setHours(0, 0, 0, 0)
      const filtDate = new Date(filters.date)
      filtDate.setHours(0, 0, 0, 0)
      return (!filters.date | matchDate.getTime() === filtDate.getTime()) &
      (!filters.place | filters.place === match.Ort) &
      (!filters.category | filters.category === match.Konkurrenz)
    })
    this.setState({ match })

    const player = { ...this.state.player }
    player.filtered = player.players
    this.setState({ player })
  }

  generateSchedule (event) {
    event.preventDefault()

    const matchlist = new Excel.Workbook()
    matchlist.SheetNames = []
    matchlist.Sheets = {}

    const worksheets = {}

    let placeArray = this.state.match.places
    if (placeArray.length > 1) {
      placeArray = placeArray.concat([FILTER_OFF])
    }
    const date = Object.keys(this.state.match.dates).find((key) =>
      String(this.state.match.dates[key]) === this.date.value
    )

    placeArray.forEach((place) => {
      let header = [
        [`Spielplan für den ${date} (${place})`],
        [],
        ['Ort', 'Zeit', 'Kategorie', 'Spieler 1', '', 'Spieler 2', '', '1. Satz', '2. Satz', '3. Satz', 'WO Grund']
      ]
      let matchListPerPlace = this.state.match.filtered.filter((match) => (match.Ort === place | place === FILTER_OFF)).map((match) =>
        [match.Ort, match.ZeitString, match.Konkurrenz, match.Spieler1, match.Spieler1Klassierung, match.Spieler2, match.Spieler2Klassierung]
      )
      console.log('Generated match list per place:', matchListPerPlace)
      worksheets[place] = Excel.SheetFromArray(header.concat(matchListPerPlace))
      matchlist.SheetNames.push(place)
      matchlist.Sheets[place] = worksheets[place]
    })

    Excel.saveAs(matchlist, 'Spielliste.xlsx')
  }

  generatePayTable (event) {
    event.preventDefault()

    const paylist = new Excel.Workbook()
    paylist.SheetNames = []
    paylist.Sheets = {}

    const worksheets = {}

    const payerList = {}
    this.state.match.matches.forEach((match) => {
      const id1 = `${match.Konkurrenz} ${match.Spieler1}`
      if (!!match.Spieler1 & !payerList.hasOwnProperty(id1)) {
        let foundPlayer = this.state.player.players.find((player) =>
          (player.name() === match.Spieler1) & (player.Konkurrenz === match.Konkurrenz)
        )
        if (!foundPlayer) {
          console.log('Debug payerlist:', foundPlayer, match)
          throw Error('Player 1 not found. This is an error!')
        }
        payerList[id1] = { ...match, Bezahlt: foundPlayer.Bezahlt }
      }

      const id2 = `${match.Konkurrenz} ${match.Spieler2}`
      if (!!match.Spieler2 & !payerList.hasOwnProperty(id2)) {
        let foundPlayer = this.state.player.players.find((player) =>
          (player.name() === match.Spieler2) & (player.Konkurrenz === match.Konkurrenz)
        )
        if (!foundPlayer) {
          console.log('Debug payerlist:', foundPlayer, match)
          throw Error('Player 2 not found. This is an error!')
        }
        payerList[id2] = { ...match, Bezahlt: foundPlayer.Bezahlt }
      }
    })
    console.log('Generated payerList', payerList)

    let placeArray = this.state.match.places
    if (placeArray.length > 1) {
      placeArray = placeArray.concat([FILTER_OFF])
    }
    const date = Object.keys(this.state.match.dates).find((key) =>
      String(this.state.match.dates[key]) === this.date.value
    )

    placeArray.forEach((place) => {
      let header = [
        [`Zahlliste für den ${date} (${place})`],
        [],
        ['bereits bez.', 'Kategorie', 'Zeit', 'Name', 'Betrag bez.', 'Quittung abgeben']
      ]

      let payListPerPlace = []
      this.state.match.filtered.forEach((match) => {
        if (!!match.Spieler1 & (match.Ort === place | FILTER_OFF === place)) {
          const id1 = `${match.Konkurrenz} ${match.Spieler1}`
          const player = payerList[id1]
          let paid = null
          if (player.Datum < this.date.value) {
            paid = player.DatumString
          }
          if (player.Bezahlt) {
            paid = 'OK'
          }
          payListPerPlace.push([ paid, match.Konkurrenz, match.ZeitString, match.Spieler1 ])
        }
        if (!!match.Spieler2 & (match.Ort === place | FILTER_OFF === place)) {
          const id2 = `${match.Konkurrenz} ${match.Spieler2}`
          const player = payerList[id2]
          let paid = null
          if (player.Datum < this.date.value) {
            paid = player.Datum
          }
          if (player.Bezahlt) {
            paid = 'OK'
          }
          payListPerPlace.push([ paid, match.Konkurrenz, match.ZeitString, match.Spieler2 ])
        }
      })
      console.log('Generated pay list per place:', payListPerPlace)
      worksheets[place] = Excel.SheetFromArray(header.concat(payListPerPlace))
      paylist.SheetNames.push(place)
      paylist.Sheets[place] = worksheets[place]
    })

    Excel.saveAs(paylist, 'Zahlliste.xlsx')
  }

  render () {
    const PlayerList = Player.components.PlayerList
    const MatchList = Match.components.MatchList

    const places = this.state.match.places
    const dates = this.state.match.dates
    const categories = this.state.match.categories
 
    return (
      <div className='App'>
        <form>
          <label htmlFor='PlayerList'>Swisstennis PlayerList.xls</label>
          <input type='file' id='PlayerList' ref={(input) => { this.playerList = input }} accept='.xls' placeholder='PlayerList File' onChange={this.handlePlayerList} />
          <label htmlFor='Calendar'>Swisstennis Calendar.xls</label>
          <input type='file' id='Calendar' ref={(input) => { this.calendar = input }} accept='.xls' placeholder='Calendar File' onChange={this.handleCalendar} />
          <label htmlFor='Date'>Datum</label>
          <select id='Date' ref={(input) => { this.date = input }} onChange={this.filterData}>
            <option>{FILTER_OFF}</option>
            {Object.keys(dates).map((key) => (
              <option key={key} value={dates[key]}>{key}</option>
            ))}
          </select>
          <label htmlFor='Place'>Ort</label>
          <select id='Place' ref={(input) => { this.place = input }} onChange={this.filterData}>
            <option>{FILTER_OFF}</option>
            {places.map((place, key) => (
              <option key={key}>{place}</option>
            ))}
          </select>
          <label htmlFor='Category'>Konkurrenz</label>
          <select id='Category' ref={(input) => { this.category = input }} onChange={this.filterData}>
            <option>{FILTER_OFF}</option>
            {categories.map((category, key) => (
              <option key={key}>{category}</option>
            ))}
          </select>
          <label htmlFor='Unpaid'>Nur unbezahlte</label>
          <input type='checkbox' id='Unpaid' ref={(input) => { this.unpaid = input }} onChange={this.filterDate} />
          <button onClick={this.generateSchedule} disabled={!this.state.match.filtered.length}>Spielliste generieren</button>
          <button onClick={this.generatePayTable} disabled={(!this.state.match.filtered.length | !this.state.player.filtered.length)}>Zahlliste generieren</button>
        </form>
        <PlayerList player={this.state.player} />
        <MatchList match={this.state.match} />
      </div>
    )
  }
}

export default App
